/*
	File:		Except.h

	Contains:	Exception handling macros

	Owned by:	Jens Alfke

	Copyright:	 1993 - 1996 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <3>	 5/24/96	jpa		1246074: Made _NATIVE_EXCEPTIONS_ work
									again. 1.1MRD: Speed/size optimizations for
									non native exceptions (_ASM_XTRY_.)
		 <2>	  3/3/96	TJ		Added global gVolatile to help make
									ODVolatile realy volatile.

	In Progress:
		
*/


/*
	THEORY OF OPERATION
	
	This is a catch/throw based exception handling package which models
	much of the behavior of native C++ exception handling. (Its macros can
	be redefined to use native exceptions if your compiler supports them.)
	
	Reading this header is not recommended for the faint of heart! Here be
	gnarly macros, setjmp fu, and tangled #ifdefs.
	Please see the recipe document "Using Exceptions" for details presented
	in a human-readable fashion.
	
	If the symbol _NATIVE_EXCEPTIONS_ is predefined, native C++ exception
	handling will be used.
	If the symbol _ASM_XTRY_ is predefined, a special assembly-coded
	version of _xTry will be called, which makes the calling code for TRY
	shorter and slightly faster. (This feature still in progress...)
*/

#ifndef _EXCEPT_
#define _EXCEPT_

#ifndef _ODTYPES_
#include "ODTypes.h"
#endif

#ifndef _ERRORDEF_
	#ifdef __cplusplus
	#include "ErrorDef.xh"	// Clients probably need the error codes as well
	#else
	#include "ErrorDef.h"
	#endif
#endif

#ifndef __SETJMP__
#include <setjmp.h>
#endif

#ifdef __LIBRARYMANAGER__
#error "Please don't include both Except.h and LibraryManager.h"
#endif


#ifdef __MWERKS__
#ifdef __cplusplus
	// The _NATIVE_EXCEPTIONS_ flag must match the Exception Support option:
	#if __option(exceptions)
		#define _NATIVE_EXCEPTIONS_
	#elif defined(_NATIVE_EXCEPTIONS_)
		#if __MWERKS__>=0x0700
			#error Must turn on "Enable Exceptions" to use Except.h with _NATIVE_EXCEPTIONS_
		#else
			#error Must turn off "direct destruction" to use Except.h with _NATIVE_EXCEPTIONS_
		#endif
	#endif
#endif
#endif


#ifdef _OD_IMPL_SHARE_UTILS_
#pragma import on
#elif defined(PRAGMA_INTERNAL_SUPPORTED)
#pragma internal on
#endif

#ifdef __cplusplus
#define EXTERNC extern "C"
#else
#define EXTERNC
#endif


//=====================================================================================
// ODVolatile
//=====================================================================================

// Any variable or parameter that is modified in a TRY block and used in the
// CATCH block must be declared as volatile or it may have an incorrect value
// in the CATCH block.
// Since not all compilers support the 'volatile' keyword, use this instead;

#ifndef _NATIVE_EXCEPTIONS_
	#if defined(__SC__) || defined(__MRC__)
		extern void* gVolatile;						/* Some compilers optimize away &x */
		#define ODVolatile(x)	(gVolatile = &x)
	#else
		#define ODVolatile(x)	((void)&x)			/* Sufficient for other compilers */
	#endif
#else
	#define ODVolatile(x)					/*no need for this with native exceptions*/
#endif

//=====================================================================================
// Exception Handling Macros (native C++ exceptions)
//=====================================================================================

#ifdef _NATIVE_EXCEPTIONS_

#define ErrorCode()				(_exception.error)
#define ErrorMessage()			(_exception.message)
#define SetErrorCode(ERR)		(_exception.error=(ERR))
#define SetErrorMessage(MSG)	(strcpy(_exception.error,(MSG))

#define TRY													\
				try {

#define CATCH_ALL											\
				} catch(ODException &_exception) {

#define RERAISE												\
				throw

#define SET_SOM_EXCEPTION(EV)								\
				ODSetSOMException(EV,_exception)

#define ENDTRY												\
		        }

/* CATCH( ) will not work with native exceptions. Don't use it! */


//=====================================================================================
// Exception Handling Macros (emulated)
//=====================================================================================

#else /*not _NATIVE_EXCEPTIONS*/

#define ErrorCode() 				(_except.fError)
#define ErrorMessage()				(_except.fODException ?_except.fODException->message :kODNULL)
#define SetErrorCode(ERR)			(_xSetErrorCode(&_except,(ERR)))
#define SetErrorMessage(MSG)		(_xSetErrorMessage(&_except,(MSG))

#ifndef _ASM_XTRY_
	#define _setjmp_xTry(EX)		setjmp(*_xTry(EX))
	#define _setjmp_xTryEv(EX,EV)	setjmp(*_xTryEv(EX,EV))
#endif

#define TRY 												\
	{														\
		ODExceptionFrame _except;							\
		if (_setjmp_xTry(&_except) == 0)					\
		{
		
#define CATCH_ALL											\
		}													\
		else												\
		{ 
		
		
#define RERAISE												\
		_xReraise(&_except)

#define SET_SOM_EXCEPTION(EV)								\
		_xSetSOMException(EV,&_except)
		
#ifdef __cplusplus	
#define ENDTRY												\
		}													\
	}
#else // C version:
#define ENDTRY												\
		}													\
		_xPop(&_except);									\
	}
#endif

// CATCH( ) is not compatible with native C++ exceptions and has been obsoleted.
#define CATCH(e)	CATCH_IS_OBSOLETE_USE_CATCH_ALL
//#define CATCH(e)											\
		} else if (_except.fError==(e)) {
		
		
#endif /*_NATIVE_EXCEPTIONS*/

//=====================================================================================
// Raising Exceptions
//=====================================================================================

EXTERNC void THROW(ODError error);
EXTERNC void THROW_IF_ERROR(ODError error);
EXTERNC void THROW_IF_NULL(void* value);

// Optional message parameters (ignored in nondebug build):
#if ODDebug
EXTERNC void THROW_IF_ERROR_M(ODError error, const char* msg);
EXTERNC void THROW_M(ODError error, const char* msg);
EXTERNC void THROW_IF_NULL_M(void* value, const char* msg);
#else
#define THROW_IF_ERROR_M(ERR,MSG)	THROW_IF_ERROR(ERR)
#define THROW_M(ERR,MSG)			THROW(ERR)
#define THROW_IF_NULL_M(ERR,MSG)	THROW_IF_NULL(ERR)
#endif

#ifdef __cplusplus
// Overloaded C++ equivalents:
inline void THROW_IF_ERROR(ODError error, const char* msg)
							{THROW_IF_ERROR_M(error,msg);}
inline void THROW(ODError error, const char* msg)
							{THROW_M(error,msg);}
inline void THROW_IF_NULL(void* value, const char* msg)
							{THROW_IF_NULL_M(value,msg);}
inline void THROW_IF_NULL(void* value, ODError error)
							{if ( value == kODNULL ) THROW(error);}
#endif

// Call BreakOnThrow(TRUE) to break into the debugger whenever THROW is called.
// (The call returns the previous value of the setting.)
EXTERNC ODBoolean BreakOnThrow( ODBoolean brk );


//=====================================================================================
// SOM Exception Utilities
//=====================================================================================

// This modified TRY block should be used in SOM methods. It's just like a
// regular TRY...CATCH_ALL...ENDTRY except that the exception code will be
// stored in the Environment. Needless to say you should _not_ reraise!
// You should also not make any SOM calls after the SOM_ENDTRY, nor declare any
// Destructo objects before the SOM_TRY.

#ifdef _NATIVE_EXCEPTIONS_

class EVSetter {	// Used to ensure ev is set at _end_ of catch block
	public:
		EVSetter( Environment* ev, ODException& x )
			:fEv(ev), fEx(x) { }
		~EVSetter( );
	private:
		Environment *fEv;
		ODException &fEx;
};

#define SOM_TRY 												\
			TRY
			
#define SOM_CATCH_ALL											\
			CATCH_ALL {											\
				EVSetter _evs(ev,_exception);					\
				
#define SOM_ENDTRY												\
			}													\
			ENDTRY

#define SOM_CATCH_ALL_ENDTRY									\
			CATCH_ALL											\
				ODSetSOMException(ev,_exception);				\
			ENDTRY												\
			// this generates more efficient code.

#else /*not _NATIVE_EXCEPTIONS_*/

#define SOM_TRY 												\
		{														\
			ODExceptionFrame _except;							\
			if (_setjmp_xTryEv(&_except,ev) == 0)				\
			{

#define SOM_CATCH_ALL											\
			CATCH_ALL
#define SOM_ENDTRY												\
			ENDTRY

#define SOM_CATCH_ALL_ENDTRY									\
			CATCH_ALL											\
			ENDTRY												\


#endif /*_NATIVE_EXCEPTIONS_*/


// ODSetSOMException stores an OD error code in the environment.
// ODGetSOMException returns the OD error code (if any) from an environment.

#ifdef __cplusplus
EXTERNC void	ODSetSOMException( Environment*, ODError, const char *msg =kODNULL );
#else
EXTERNC void	ODSetSOMException( Environment*, ODError, const char *msg );
#endif

#ifdef _NATIVE_EXCEPTIONS_
void	ODSetSOMException( Environment*, ODException& );
#endif

EXTERNC ODError	ODGetSOMException( Environment *ev );

// CHECK_ENV throws an exception if the environment indicates an error.

EXTERNC void	CHECK_ENV( Environment* );

// SOMCHKEXCEPT is a macro that is called in a .xh file if the ev variable
// indicates an exception is set.
#define SOMCHKEXCEPT {CHECK_ENV(ev);}


//=====================================================================================
// Obsolete Exception Utilities
//=====================================================================================

/*  FN_CATCH and SOM_CATCH are now obsolete. Please convert to regular TRY or SOM_TRY. */
#if 0

#if defined(__MWERKS__) || defined(__SC__)
	extern void FN_CATCH_DIDNT_RETURN( );
	/* This function is purposely declared but never defined. Due to the setup of
	   the for loop in FN_CATCH and SOM_CATCH, a failure to end the block with a
	   return statement will cause the call to FN_CATCH_DIDNT_RETURN not to be
	   dead stripped, resulting in a link error. This will help catch this nasty
	   coding error. */
#else
	// Other compilers may not dead-strip properly.
	#define FN_CATCH_DIDNT_RETURN() /**/
#endif

#define FN_CATCH											\
		ODExceptionFrame _except;							\
		ODVolatile(_except);								\
		if (_setjmp_xTry(&_except) != 0)					\
			for( ; true; FN_CATCH_DIDNT_RETURN() )

#define SOM_CATCH											\
		ODExceptionFrame _except;							\
		ODVolatile(_except);								\
		if (_setjmp_xTryEv(&_except,ev) != 0)				\
			for( ; true; FN_CATCH_DIDNT_RETURN() )

#endif /*0*/

//=====================================================================================
// Finally, the Exception Handler (ODExceptionFrame struct)
//=====================================================================================


#ifndef _NATIVE_EXCEPTIONS_

#ifdef __cplusplus
class Destructo;	// forward declarations for use below
struct ODExceptionFrame;
EXTERNC void _xPop(ODExceptionFrame*);
#else
typedef struct ODExceptionFrame ODExceptionFrame;
typedef struct Destructo Destructo;		// Not used in C but must declare anyway
#endif

struct ODExceptionFrame
{
	ODExceptionFrame*	fPrev;
	ODException*		fODException;
	ODError				fError;
	Environment*		fEv;
	Destructo*			fDestructoList;
	jmp_buf				fBuffer;
#ifdef __cplusplus
	void Throw( ODError err, const char* msg, ODException *x =kODNULL );
	inline ~ODExceptionFrame( )	{_xPop(this);}
#endif
};

typedef struct ODExceptionFrame ODExceptionFrame;

#ifdef _ASM_XTRY_
	EXTERNC int _setjmp_xTry( ODExceptionFrame* );
	EXTERNC int _setjmp_xTryEv( ODExceptionFrame*, Environment* );
#else
	EXTERNC jmp_buf* _xTry( ODExceptionFrame* );
	EXTERNC jmp_buf* _xTryEv( ODExceptionFrame*, Environment* );
#endif

EXTERNC void _xReraise( ODExceptionFrame* );
EXTERNC void _xSetSOMException(Environment*,ODExceptionFrame*);
EXTERNC void _xPop( ODExceptionFrame* );
EXTERNC void _xSetErrorCode( ODExceptionFrame*, ODError );
EXTERNC void _xSetErrorMessage( ODExceptionFrame*, const char* );

#endif /*_NATIVE_EXCEPTIONS_*/


//=====================================================================================
// Destructo, a C++ base class for auto-destruct objects
//=====================================================================================

#ifdef __cplusplus

class Destructo
{
#ifndef _NATIVE_EXCEPTIONS_
protected:
	Destructo( );
	
public:
	virtual ~Destructo( );
	
private:
	Destructo* EmergencyDestruct( );
	Destructo *fPrevDestructo;
	friend class ODExceptionFrame;

	static void* operator new( size_t );	// Make it illegal to allocate on heap
											// In future could use alloca??
											
	private: // disallow these:
	Destructo(const Destructo& );
	void operator=(const Destructo& );
	// Bitwise assigning one destructo to another smashes
	// fPrevDestructo with potentially unpleasant effects.
#endif /*_NATIVE_EXCEPTIONS_*/
};

#endif /*__cplusplus*/

#ifdef _OD_IMPL_SHARE_UTILS_
#pragma import off
#elif defined(PRAGMA_INTERNAL_SUPPORTED)
#pragma internal off
#endif


#endif // _EXCEPT_
